@@ -1,92 +1,11 @@
+import functools
 import gevent
+from gevent import event, pool
 import logbook
 
 from . import config as biloba_config, events
 
 
-class ThreadPool(object):
-    """
-    Create and maintain a group of unrelated greenlets.
-
-    Spawn threads by calling ``spawn`` in the same was as you would create a
-    greenlet. Kill the pool by calling ``kill``.
-
-    :ivar threads: A list of thread maintained by this pool.
-    :ivar dead: Whether this pool is dead/dying.
-    """
-
-    __slots__ = (
-        'threads',
-        'dead',
-    )
-
-    def __init__(self):
-        self.threads = []
-        self.dead = False
-
-    def kill(self):
-        """
-        Kill all the threads in this pool and signal that the pool is dead
-        """
-        threads, self.threads = self.threads, []
-        self.dead = True
-
-        gevent.killall(threads)
-
-        self.join()
-
-    def join(self, timeout=None):
-        """
-        Wait until there are either no threads in the pool, or the pool has
-        been killed.
-
-        :param timeout: Number of seconds to wait before ``gevent.Timeout`` is
-            raised. By default, there is no timeout.
-        """
-        while not self.dead and self.threads:
-            gevent.joinall(self.threads, timeout=timeout)
-
-    def remove(self, thread):
-        """
-        Remove a thread from the pool.
-
-        :param thread: The thread to be removed.
-        """
-        try:
-            self.threads.remove(thread)
-        except ValueError:
-            pass
-
-    def spawn(self, func, *args, **kwargs):
-        """
-        Spawns a greenlet within this pool and will be removed when it 'ready'.
-
-        :param func: The callable to execute in a new greenlet context.
-        :param args: The args to pass to the callable.
-        :param kwargs: The kwargs to pass to the callable.
-        :returns: The spawned greenlet thread.
-        """
-        if self.dead:
-            raise RuntimeError('this pool is dead')
-
-        thread = gevent.spawn(func, *args, **kwargs)
-
-        self.add(thread)
-
-        return thread
-
-    def add(self, thread):
-        """
-        Add an existing thread to this pool.
-        """
-        if self.dead:
-            raise RuntimeError('this pool is dead')
-
-        self.threads.append(thread)
-
-        thread.rawlink(self.remove)
-
-
 class Service(events.EventEmitter):
     """
     An asynchronous primitive that will maintain a pool of spawned greenlets
@@ -109,6 +28,8 @@ class Service(events.EventEmitter):
         watching.
     :ivar pool: A thread pool controlled by this service. If the threadpool
         empties, this service is dead.
+    :ivar logger: The logbook instance that is used by the service to log
+        interesting events.
     """
 
     __slots__ = (
@@ -116,24 +37,22 @@ class Service(events.EventEmitter):
         'services',
         'pool',
         'logger',
+        '_run_thread',
+        '_kill',
     )
 
     # set to specify the logger name (before the first access)
     logger_name = None
 
-    def __init__(self):
+    def __init__(self, logger=None):
         super(Service, self).__init__()
 
         self.started = False
         self.services = []
-        self.pool = ThreadPool()
-        self.logger = self.get_logger()
-
-    def __del__(self):
-        try:
-            self.stop()
-        except:
-            pass
+        self.pool = pool.Group()
+        self.logger = logger or self.get_logger()
+        self._run_thread = None
+        self._kill = event.Event()
 
     def get_logger(self):
         return logbook.Logger(self.logger_name or self.__class__.__name__)
@@ -155,7 +74,7 @@ class Service(events.EventEmitter):
         Perform any necessary cleanup.
         """
 
-    def start(self):
+    def start(self, block=True):
         """
         Called to start this service and any child services that may be
         registered.
@@ -165,25 +84,49 @@ class Service(events.EventEmitter):
         if self.started:
             return
 
-        with self.emit_exceptions():
-            try:
-                self.do_start()
+        if not self._run_thread:
+            # there is no running thread, let's start it
+            self._run_thread = gevent.spawn(self._run)
 
-                for child in self.services:
-                    self.spawn(self.watch_service, child)
-            except:
-                self.teardown()
+            # ensure that the _run_thread attribute is cleaned up when the
+            # greenlet comes to an end
+            def cleanup(g):
+                self._run_thread = None
 
-                raise
+            self._run_thread.rawlink(cleanup)
 
-        self.started = True
+        if not block:
+            return
 
-        try:
-            self.emit('start')
-        except:
-            self.stop()
+        # block this greenlet until the run thread starts the service
+        result = gevent.event.AsyncResult()
 
-            raise
+        @self.once('error')
+        def on_error(*exc_info):
+            result.set_exception(exc_info[1])
+            self.remove_listener('start', on_start)
+
+        @self.once('start')
+        def on_start():
+            result.set()
+            self.remove_listener('error', on_error)
+
+        result.get()
+
+    def stop(self, block=True):
+        """
+        Called to stop this service if it is running. All child services are
+        `stopped` and any greenlets this service may have spawned are killed.
+
+        Any custom code should generally go in to `do_stop`.
+        """
+        if not self.started:
+            return
+
+        self._kill.set()
+
+        if block:
+            self._run_thread.get()
 
     def teardown(self):
         """
@@ -194,29 +137,51 @@ class Service(events.EventEmitter):
         for service in self.services:
             service.stop()
 
-        self.pool.kill()
         self.services = []
+        self.pool.kill()
 
-    def stop(self):
+    def start_service(self):
         """
-        Called to stop this service if it is running. All child services are
-        `stopped` and any greenlets this service may have spawned are killed.
+        Called in the running service thread to start the service. This method
+        blocks until all the child services have started.
+        """
+        self.do_start()
 
-        Any custom code should generally go in to `do_stop`.
-        """
-        if not self.started:
-            return
+        for child in self.services:
+            child.start(block=True)
+
+            self.spawn(self.watch_service, child)
+
+        self.started = True
 
         try:
-            with self.emit_exceptions():
-                self.teardown()
-                self.do_stop()
-        finally:
-            self.started = False
+            self.emit('start')
+        except:
+            self.stop()
 
-        # it is important that everything is torn down before the event is
-        # emitted
-        self.emit('stop')
+            raise
+
+    def stop_service(self):
+        try:
+            self.teardown()
+        finally:
+            try:
+                self.do_stop()
+            finally:
+                self.started = False
+
+                # it is important that everything is torn down before the
+                # event is emitted
+                self.emit('stop')
+
+    def _run(self):
+        with self.emit_exceptions(propagate=False):
+            try:
+                self.start_service()
+
+                self._kill.wait()
+            finally:
+                self.stop_service()
 
     def join(self):
         """
@@ -225,9 +190,7 @@ class Service(events.EventEmitter):
         # start is idempotent
         self.start()
 
-        self.pool.join()
-
-        self.stop()
+        self._run_thread.get()
 
     def handle_service_error(self, service, *exc_info):
         """
@@ -279,22 +242,9 @@ class Service(events.EventEmitter):
             propagate=propagate,
             always_log=always_log,
             emit=emit,
-            skip_types=(gevent.GreenletExit,)
+            skip_types=(gevent.GreenletExit,),
         )
 
-    def exec_func(self, func, *args, **kwargs):
-        """
-        Execute the function and emit an error if an exception occurs. If
-        the error is not trapped/handled, the default behaviour is to
-        output the exception to the logger.
-
-        :param func: The function to execute.
-        :param args: The params to pass to `func`.
-        :param kwargs: The keyword arguments to pass to `func`.
-        """
-        with self.emit_exceptions(propagate=False):
-            return func(*args, **kwargs)
-
     def spawn(self, func, *args, **kwargs):
         """
         Spawns a greenlet that is linked to this service and will be killed if
@@ -305,7 +255,12 @@ class Service(events.EventEmitter):
         :param kwargs: The kwargs to pass to the callable.
         :returns: The spawned greenlet thread.
         """
-        return self.pool.spawn(self.exec_func, func, *args, **kwargs)
+        @functools.wraps(func)
+        def wrapped():
+            with self.emit_exceptions(propagate=False):
+                return func(*args, **kwargs)
+
+        return self.pool.spawn(wrapped)
 
     def watch_service(self, child):
         """
@@ -315,7 +270,7 @@ class Service(events.EventEmitter):
             child.join()
             self.logger.debug('{!r} completed'.format(child))
         finally:
-            self.teardown()
+            self.stop()
 
 
 class ConfigurableService(Service):
